/*
 * Copyright 2006 - 2011 
 *     Julien Baudry	<julien.baudry@graphstream-project.org>
 *     Antoine Dutot	<antoine.dutot@graphstream-project.org>
 *     Yoann Pigné		<yoann.pigne@graphstream-project.org>
 *     Guilhelm Savin	<guilhelm.savin@graphstream-project.org>
 * 
 * This file is part of GraphStream <http://graphstream-project.org>.
 * 
 * GraphStream is a library whose purpose is to handle static or dynamic
 * graph, create them from scratch, file or any source and display them.
 * 
 * This program is free software distributed under the terms of two licenses, the
 * CeCILL-C license that fits European law, and the GNU Lesser General Public
 * License. You can  use, modify and/ or redistribute the software under the terms
 * of the CeCILL-C license as circulated by CEA, CNRS and INRIA at the following
 * URL <http://www.cecill.info> or under the terms of the GNU LGPL as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-C and LGPL licenses and that you accept their terms.
 */
package org.graphstream.graph.implementations;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;

import org.graphstream.graph.Edge;
import org.graphstream.graph.Graph;
import org.graphstream.graph.Node;
import org.graphstream.graph.IdAlreadyInUseException;

/**
 * Full implementation of {@link org.graphstream.graph.Node} that allows only
 * one edge between two nodes.
 */
public class SingleNode extends DefaultNode {
	// Attribute

	/**
	 * Map of leaving edges toward nodes. Each element of the map is a pair
	 * (key,value) where the key is the id of a node that can be reached
	 * following a leaving edge, and the value is the leaving edge.
	 */
	protected HashMap<String, Edge> to = new HashMap<String, Edge>();

	/**
	 * Map of entering edges from nodes. Each element of the map is a pair
	 * (key,value) where the key is the id of a node that can be reached
	 * following an entering edge, and the value is the entering edge.
	 */
	protected HashMap<String, Edge> from = new HashMap<String, Edge>();

	// Constructor

	/**
	 * New unconnected node.
	 * 
	 * @param graph
	 *            The graph containing the node.
	 * @param id
	 *            Tag of the node.
	 */
	public SingleNode(Graph graph, String id) {
		super(graph, id);
	}

	// Access

	@Override
	public int getOutDegree() {
		return to.size();
	}

	@Override
	public int getInDegree() {
		return from.size();
	}

	@Override
	public boolean hasEdgeToward(String id) {
		return (to.get(id) != null);
	}

	@Override
	public boolean hasEdgeFrom(String id) {
		return (from.get(id) != null);
	}

	@Override
	@SuppressWarnings("unchecked")
	public <T extends Edge> T getEdgeToward(String id) {
		return (T) to.get(id);
	}

	@Override
	@SuppressWarnings("unchecked")
	public <T extends Edge> T getEdgeFrom(String id) {
		return (T) from.get(id);
	}

	@Override
	public <T extends Edge> T getEdgeBetween(String id) {
		if (hasEdgeToward(id))
			return getEdgeToward(id);
		else
			return getEdgeFrom(id);
	}

	@Override
	@SuppressWarnings("unchecked")
	public <T extends Edge> Iterator<T> getEnteringEdgeIterator() {
		return new ElementIterator<T>((HashMap<String, T>) from);
	}

	@Override
	@SuppressWarnings("unchecked")
	public <T extends Edge> Iterator<T> getLeavingEdgeIterator() {
		return new ElementIterator<T>((HashMap<String, T>) to);
	}

	// Access -- Not in Node interface

	@Override
	@SuppressWarnings("unchecked")
	public <T extends Edge> Iterable<T> getEachLeavingEdge() {
		return (Iterable<T>) to.values();
	}

	@Override
	@SuppressWarnings("unchecked")
	public <T extends Edge> Iterable<T> getEachEnteringEdge() {
		return (Iterable<T>) from.values();
	}

	@Override
	@SuppressWarnings("unchecked")
	public <T extends Edge> Collection<T> getLeavingEdgeSet() {
		return (Collection<T>) Collections.unmodifiableCollection(to.values());
	}

	@Override
	@SuppressWarnings("unchecked")
	public <T extends Edge> Collection<T> getEnteringEdgeSet() {
		return (Collection<T>) Collections
				.unmodifiableCollection(from.values());
	}

	// Command

	/**
	 * Add an edge between this node and the given target.
	 * 
	 * @param tag
	 *            Tag of the edge.
	 * @param target
	 *            Target node.
	 * @param directed
	 *            If the edge is directed only from this node to the target.
	 */
	@Override
	@SuppressWarnings("unchecked")
	protected <T extends Edge> T addEdgeToward(String tag, DefaultNode target,
			boolean directed) throws IllegalArgumentException {
		// Some checks.

		if (target.G == null)
			throw new IllegalArgumentException("cannot add edge to node `"
					+ target.getId()
					+ "' since this node is not yet part of a graph");

		if (G == null)
			throw new IllegalArgumentException("cannot add edge to node `"
					+ getId() + "' since this node is not yet part of a graph");

		if (G != target.G)
			throw new IllegalArgumentException("cannot add edge between node `"
					+ getId() + "' and node `" + target.getId()
					+ "' since they pertain to distinct graphs");

		// Register the edge.

		Edge edge = target.getEdgeToward(getId());

		if (edge != null) {
			throw new IllegalArgumentException(new IdAlreadyInUseException(tag));
		} else {
			T e = (T) G.edgeFactory.newInstance(tag, this, target, directed);
			// e.setDirected(directed);
			// return new CheckedEdge( tag, this, target, directed );
			return e;
		}
	}

	/**
	 * Called by an edge to bind it.
	 */
	@Override
	protected void registerEdge(Edge edge) throws IllegalArgumentException,
			IdAlreadyInUseException {
		// If the edge or an edge with the same id is already registered.

		Node other = edge.getOpposite(this);

		if (other != this) // case of loop edges
		{
			if (getEdgeToward((other).getId()) != null
					|| getEdgeFrom((other).getId()) != null)
				throw new IdAlreadyInUseException(
						"multi edges are not supported: edge between node '"
								+ getId() + "' and '" + (other).getId()
								+ "' already exists");
		}

		// Add the edge.

		edges.add(edge);

		String otherId = other.getId();

		if (edge.isDirected()) {
			if (edge.getSourceNode() == this)
				to.put(otherId, edge);
			else
				from.put(otherId, edge);
		} else {

			to.put(otherId, edge);
			from.put(otherId, edge);
		}
	}

	@Override
	protected void unregisterEdge(Edge edge) {
		Node other = edge.getOpposite(this);

		to.remove(other.getId());
		from.remove(other.getId());

		int pos = edges.indexOf(edge);

		if (pos >= 0)
			edges.remove(pos);
	}

	/**
	 * When a node is unregistered from a graph, it must not keep edges
	 * connected to nodes still in the graph. This methods untie all edges
	 * connected to this node (this also unregister them from the graph).
	 */
	@Override
	protected void disconnectAllEdges() throws IllegalStateException {
		int n = edges.size();

		// We cannot use a "for" since untying an edge removes this edge from
		// the node. The number of edges will change continuously.

		while (n > 0) {
			Edge e = edges.get(0);
			G.removeEdge(((AbstractElement) e).getId());
			// e.unbind();
			n = edges.size();
		}
	}
}